import argparse
import cookiecutter.config as cc_config
from cliff.command import Command

from .command_helpers import CommandHelpers
from .help_text import (
    NO_TEMPLATES_ERROR_WITH_HELP_MSG,
    TEMPLATE_USAGE_MSG,
    TEMPLATE_UPDATE_MSG
)
from datakit_project.cookiecutters import Cookiecutters
from datakit_project.formatters.templates import Templates as formatter


class Templates(CommandHelpers, Command):
    "List project templates"

    def get_parser(self, prog_name):
        parser = super(Templates, self).get_parser(prog_name)
        parser.formatter_class = argparse.RawTextHelpFormatter
        parser.add_argument(
            '-s',
            '--status',
            action='store_true',
            default=False,
            help="Check for updates on local Cookiecutter templates"
        )
        return parser

    def take_action(self, parsed_args):
        cc_home = cc_config.DEFAULT_CONFIG['cookiecutters_dir']
        cookiecutters = Cookiecutters(cc_home)
        templates = cookiecutters.info(status=parsed_args.status)
        self.log.info("\nLocal Cookiecutter templates ({}):\n".format(cc_home))
        if len(templates) == 0:
            self.log.info(NO_TEMPLATES_ERROR_WITH_HELP_MSG)
        else:
            default_cc_name = self.configs["default_template"].split(cc_home)[1]
            if parsed_args.status:
                tbl = formatter.status(templates, default_cc_name)
            else:
                tbl = formatter.list(templates, default_cc_name)
            self.log.info(tbl)
            self.log.info("  *: Default template")
            self.log.info('\n')
            self.log.info(TEMPLATE_USAGE_MSG)
            self.log.info(TEMPLATE_UPDATE_MSG)



import os
import shutil

import cookiecutter.config as cc_config
import pytest
from datakit.utils import makedirs, mkdir_p, write_json


@pytest.fixture
def cookiecutter_home(tmpdir):
    return os.path.join(str(tmpdir), '.cookiecutters')


@pytest.fixture
def datakit_home(tmpdir):
    return os.path.join(str(tmpdir), '.datakit')


@pytest.fixture
def plugin_dir(datakit_home):
    return os.path.join(datakit_home, 'plugins/datakit-project')


@pytest.fixture
def create_plugin_config_default(plugin_dir):
    config_file = os.path.join(plugin_dir, 'config.json')
    config = {'default_template': ''}
    write_json(config_file, config)


@pytest.fixture
def create_plugin_config_fake_repo(plugin_dir):
    config_file = os.path.join(plugin_dir, 'config.json')
    fake_repo_path = os.path.join(os.getcwd(), 'tests/fake-repo')
    config = {'default_template': fake_repo_path}
    write_json(config_file, config)


@pytest.fixture
def deploy_template():
    copied_repos = []
    def _deploy_template(cc_home, repo_path):
        src_repo = repo_path
        base_name = repo_path.split('/')[-1]
        dest_repo = os.path.join(cc_home, base_name)
        shutil.copytree(src_repo, dest_repo)
        copied_repos.append(dest_repo)
        return dest_repo

    yield _deploy_template

    for repo in copied_repos:
        shutil.rmtree(repo)


@pytest.fixture(autouse='session')
def setup_environment(monkeypatch, cookiecutter_home, datakit_home):
    mkdir_p(cookiecutter_home)
    mkdir_p(datakit_home)
    monkeypatch.setitem(
        cc_config.DEFAULT_CONFIG,
        'cookiecutters_dir',
        cookiecutter_home,
    )
    monkeypatch.setenv('DATAKIT_HOME', datakit_home)

@pytest.fixture(autouse="module")
def create_plugin_dir(plugin_dir):
    mkdir_p(plugin_dir)
    yield
    shutil.rmtree(plugin_dir)



import os


import cookiecutter.config as cc_config
import pytest

from datakit_project.utils import resolve_repo_dir


# TODO: Update resolve_repo_dir to use cookiecutter DEFAULT_CONFIG
# then monkeypatch the variable here
def test_repo_dir_for_local_repo():
    """
    Should be fully-qualified path to local directory
    """
    local_dir = '/Local/path/to/fake-repo'
    actual_dir = resolve_repo_dir(local_dir)
    assert local_dir == actual_dir


def test_repo_dir_for_alias():
    """
    Should be path inside of cookiecutter's dir.
    """
    cc_home = cc_config.DEFAULT_CONFIG['cookiecutters_dir']
    expected_dir = os.path.join(cc_home, 'fake-repo')
    actual_dir = resolve_repo_dir('gh:associatedpress/fake-repo')
    assert expected_dir == actual_dir


def test_repo_dir_for_url():
    """
    Should be path inside of cookiecutter's dir.
    """
    cc_home = cc_config.DEFAULT_CONFIG['cookiecutters_dir']
    expected_dir = os.path.join(cc_home, 'fake-repo')
    actual_dir = resolve_repo_dir('https://github.com/associatedpress/fake-repo.git')
    assert expected_dir == actual_dir



import os
import re

from cookiecutter.config import BUILTIN_ABBREVIATIONS, DEFAULT_CONFIG
from cookiecutter.repository import expand_abbreviations, is_repo_url
from cookiecutter.vcs import identify_repo


def resolve_repo_dir(template):
    """
    Determine path to local repository based on template name.
    This can be either a local path, a remote URL or a remote alias
    such as for Github.

    Sundry bits from Cookiecutter do the heavy lifting here.

    Arguments

      template [path|url|alias]

    """
    cc_home = DEFAULT_CONFIG['cookiecutters_dir']
    if template_type(template) in ['remote alias', 'url']:
        raw_url = expand_abbreviations(template, BUILTIN_ABBREVIATIONS)
        repo_type, repo_url = identify_repo(raw_url)
        repo_url = repo_url.rstrip('/')
        tail = os.path.split(repo_url)[1]
        if repo_type == 'git':
            repo_dir = os.path.normpath(os.path.join(cc_home,
                                                     tail.rsplit('.git')[0]))
        elif repo_type == 'hg':
            repo_dir = os.path.normpath(os.path.join(cc_home, tail))
            repo_dir = template
    else:
        repo_dir = os.path.abspath(template)
    return repo_dir


def template_type(template):
    status = 'local directory'
    if re.match(r'(gh|hg):', template):
        status = 'remote alias'
    if is_repo_url(template):
        status = 'url'
    return status



import argparse
import cookiecutter.config as cc_config
from cliff.command import Command

from .command_helpers import CommandHelpers
from .help_text import TEMPLATE_STATUS_MSG
from datakit_project.cookiecutters import Cookiecutters
from datakit_project.prompt import read_multichoice_or_all_input


class TemplatesUpdate(CommandHelpers, Command):
    "Update locally installed Cookiecutter templates"

    def get_parser(self, prog_name):
        parser = super(TemplatesUpdate, self).get_parser(prog_name)
        parser.formatter_class = argparse.RawTextHelpFormatter
        return parser

    def take_action(self, parsed_args):
        cc_home = cc_config.DEFAULT_CONFIG['cookiecutters_dir']
        cookiecutters = Cookiecutters(cc_home)
        templates = cookiecutters.info(status=True)
        out_of_date = [t for t in templates if t['commits_behind'] != 'Up-to-date']
        if len(out_of_date) == 0:
            self.log.info("Local templates are all up-to-date.\n")
            self.log.info(TEMPLATE_STATUS_MSG)
        else:
            self.log.info("The following template(s) are out of date:\n")
            choices = {}
            for idx, template in enumerate(out_of_date):
                option_num = str(idx + 1)
                choices[option_num] = template
                self.log.info("({}) {}\n".format(option_num, template['Name']))
            prompt_msg = (
                "Type one or more numbers for the templates you'd like to update, "
                "or type 'all'.\n"
            )
            response = read_multichoice_or_all_input(prompt_msg)
            if 'all' in response:
                to_update = sorted([temp['Name'] for temp in choices.values()])
                self.log.info("Updating all templates...")
            else:
                to_update = [choices[num]['Name'] for num in response]
                self.log.info("Updating selected templates...")
            cookiecutters.update(to_update)



import os
import re
import shutil
import subprocess
from unittest import mock

import pytest

from datakit_project import Create



#TODO: use parametrize with repo data for Cookiecutters
@pytest.mark.usefixtures('create_plugin_config_fake_repo')
def test_interactive_project_selection(caplog, cookiecutter_home, deploy_template, monkeypatch, tmpdir):
    deploy_template(cookiecutter_home, 'tests/fake-repo')
    deploy_template(cookiecutter_home, 'tests/fake-repo-two')
    # Set up the configs to point default template at our fake repo
    monkeypatch.chdir(tmpdir)
    with mock.patch('datakit_project.commands.create.Cookiecutters') as MockClass:
        instance = MockClass.return_value
        instance.list_templates.return_value = ['fake-repo', 'fake-repo-two']
        # Monkeypatch user selection
        monkeypatch.setattr(
            'datakit_project.commands.create.read_user_choice',
            lambda name, choices: 'fake-repo'
        )
        parsed_args = mock.Mock()
        parsed_args.make_default = False
        parsed_args.template = ''
        parsed_args.interactive = True
        # Use no_input to disable additional cookiecutter prompts
        parsed_args.no_input = True
        cmd = Create(None, None, cmd_name='project create')
        cmd.run(parsed_args)
        assert "Creating project from template: fake-repo" in caplog.text



import os
import re
import shutil
from unittest import mock

import cookiecutter.config as cc_config
import pytest

from datakit_project import cookiecutters, Templates


def test_no_templates(caplog, cookiecutter_home, monkeypatch, tmpdir):
    """
    Templates should provide helpful info if no cookiecutter templates are installed
    """
    # Switch directories
    monkeypatch.chdir(tmpdir)
    # Run the command
    parsed_args = mock.Mock()
    cmd = Templates(None, None, cmd_name='project templates')
    cmd.run(parsed_args)
    assert cookiecutter_home in caplog.text
    assert "No project templates have been installed!" in caplog.text


def test_multiple_templates(caplog, cookiecutter_home, deploy_template, monkeypatch, tmpdir):
    """
    Templates should list installed cookiecutters
    """
    deploy_template(cookiecutter_home, 'tests/fake-repo')
    deploy_template(cookiecutter_home, 'tests/fake-repo-two')
    # Switch directories
    monkeypatch.chdir(tmpdir)
    # Mock the return value for Cookiecutters.info
    with mock.patch('datakit_project.commands.templates.Cookiecutters') as MockClass:
        instance = MockClass.return_value
        instance.info.return_value = [
            {
              'Name': 'fake-repo',
              'SHA': 'a11706f',
              'Date': '2017-02-02',
              'Subject': 'Testing'
            },
            {
              'Name': 'fake-repo-two',
              'SHA': '55f62a0',
              'Date': '2019-03-06',
              'Subject': 'Testing'
            },
        ]
        # Run the command
        parsed_args = mock.Mock()
        parsed_args.status = False
        cmd = Templates(None, None, cmd_name='project templates')
        cmd.run(parsed_args)
        header = "Name            SHA       Date         Subject"
        repo1 = "fake-repo       a11706f   2017-02-02   Testing"
        repo2 = "fake-repo-two   55f62a0   2019-03-06   Testing"
        expected = [header, repo1, repo2]
        # Some gross cleanup of caplog.text is necessary because it displays log level info
        actual = [re.sub(r"templates.py\s+\d+\sINFO", "", line).strip()  for line in caplog.text.split('\n')]
        # Test that expected lines appear in displayed text
        for line in expected:
            assert line in actual
        # Test alpha sortting by repo name
        header_idx = actual.index(header)
        repo1_idx = actual.index(repo1)
        repo2_idx = actual.index(repo2)
        assert header_idx < repo1_idx
        assert repo1_idx < repo2_idx



import os
import re
import shutil
from unittest import mock

import cookiecutter.config as cc_config
import pytest

from datakit_project import cookiecutters, TemplatesUpdate


def test_nothing_to_update(caplog):
    with mock.patch('datakit_project.commands.templates_update.Cookiecutters') as MockClass:
        instance = MockClass.return_value
        instance.info.return_value = [{
          'Name': 'fake-repo', 'SHA': 'a11706f', 'Date': '2017-02-02',
          'Subject': 'Testing', 'upstream_sha': 'a11706f',
          'upstream_date': '2017-02-02', 'upstream_subject': 'Testing',
          'commits_behind': 'Up-to-date'
        }]
        instance.upstream_tracking_branch.return_value = 'origin/master'
        # Run the command
        parsed_args = mock.Mock()
        cmd = TemplatesUpdate(None, None, cmd_name='project templates update')
        cmd.run(parsed_args)
        instance.info.assert_called_once_with(status=True)
        assert "Local templates are all up-to-date" in caplog.text

@pytest.mark.parametrize("repo_info, selection, expected",
    [
        # Select "all" out-of-date repos
        (
            [{
              'Name': 'fake-repo', 'SHA': 'a11706f', 'Date': '2017-02-02',
              'Subject': 'Testing', 'upstream_sha': 'c18705e',
              'upstream_date': '2017-06-01', 'upstream_subject': 'Some newer commit',
              'commits_behind': 3
            },
            {
              'Name': 'fake-repo-two', 'SHA': '55f62a0', 'Date': '2019-03-06',
              'Subject': 'Testing', 'upstream_sha': '55f62a0',
              'upstream_date': '2019-03-06', 'upstream_subject': 'Testing',
              'commits_behind': 'Up-to-date'
            }],
            ['all'],
            ['fake-repo']
        ),
        # Select single option
        (
            [{
              'Name': 'fake-repo', 'SHA': 'a11706f', 'Date': '2017-02-02',
              'Subject': 'Testing', 'upstream_sha': 'c18705e',
              'upstream_date': '2017-02-02', 'upstream_subject': 'Some newer commit',
              'commits_behind': 3
            },
            {
              'Name': 'fake-repo-two', 'SHA': '55f62a0', 'Date': '2019-03-06',
              'Subject': 'Testing', 'upstream_sha': '55f62a0',
              'upstream_date': '2019-03-06', 'upstream_subject': 'Testing',
              'commits_behind': 'Up-to-date'
            }],
            ['1'],
            ['fake-repo']
        ),
        # Select subset of out-of-date repos
        (
            [{
              'Name': 'fake-repo', 'SHA': 'a11706f', 'Date': '2017-02-02',
              'Subject': 'Testing', 'upstream_sha': 'c18705e',
              'upstream_date': '2017-02-02', 'upstream_subject': 'Some newer commit',
              'commits_behind': 3
            },
            {
              'Name': 'fake-repo-two', 'SHA': '55f62a0', 'Date': '2019-03-06',
              'Subject': 'Testing', 'upstream_sha': '32a42b0',
              'upstream_date': '2019-03-07', 'upstream_subject': 'A newer commit',
              'commits_behind': 1
            }],
            ['1'],
            ['fake-repo']
        ),
        # Select multiple out-of-date repos
        (
            [{
              'Name': 'fake-repo', 'SHA': 'a11706f', 'Date': '2017-02-02',
              'Subject': 'Testing', 'upstream_sha': 'c18705e',
              'upstream_date': '2017-02-02', 'upstream_subject': 'Some newer commit',
              'commits_behind': 3
            },
            {
              'Name': 'fake-repo-two', 'SHA': '55f62a0', 'Date': '2019-03-06',
              'Subject': 'Testing', 'upstream_sha': '32a42b0',
              'upstream_date': '2019-03-07', 'upstream_subject': 'A newer commit',
              'commits_behind': 1
            }],
            ['1', '2'],
            ['fake-repo', 'fake-repo-two']
        ),
    ]
)
def test_update(repo_info, selection, expected, monkeypatch):
    with mock.patch('datakit_project.commands.templates_update.Cookiecutters') as MockClass:
        instance = MockClass.return_value
        instance.info.return_value = repo_info
        instance.upstream_tracking_branch.return_value = 'origin/master'
        monkeypatch.setattr(
            'datakit_project.commands.templates_update.read_multichoice_or_all_input',
            lambda choices: selection
        )
        # Run the command
        parsed_args = mock.Mock()
        cmd = TemplatesUpdate(None, None, cmd_name='project templates update')
        cmd.run(parsed_args)
        instance.info.assert_called_once_with(status=True)
        instance.update.assert_called_once_with(expected)



import os

from .repository import Repository


class Cookiecutters:

    def __init__(self, cookiecutters_dir):
        self.cookiecutters_dir = cookiecutters_dir

    def info(self, status=False):
        repos_info = []
        for template in self.list_templates():
            repo_dir = os.path.join(self.cookiecutters_dir, template)
            repo = Repository(repo_dir)
            try:
                data = repo.info()
            except NotADirectoryError:
                continue
            except AttributeError as e:
                continue
            if status:
                repo.fetch()
                upstream = repo.upstream_info()
                if upstream['commits_behind'] == 0:
                    upstream['commits_behind'] = 'Up-to-date'
                data.update(upstream)
            repos_info.append(data)
        return repos_info

    def update(self, templates=[]):
        for template in templates:
            repo_dir = os.path.join(self.cookiecutters_dir, template)
            repo = Repository(repo_dir)
            repo.update()

    def list_templates(self):
        return os.listdir(self.cookiecutters_dir)



import prettytable as pt


class Templates:

    @classmethod
    def list(cls, data, default_cc_name):
        fields = ('Name', 'SHA', 'Date', 'Subject')
        tbl = cls._ptable(fields)
        for row in data:
            if (row['Name'] == default_cc_name):
                row['Name'] += '*'
            vals = (
                row['Name'],
                row['SHA'],
                row['Date'],
                row['Subject']
            )
            tbl.add_row(vals)
        return tbl

    @classmethod
    def status(cls, data, default_cc_name):
        fields = (
            'Name',
            'SHA',
            'Date',
            'Upstream SHA',
            'Upstream Date',
            'Commits behind',
        )
        tbl = cls._ptable(fields)
        for row in data:
            if (row['Name'] == default_cc_name):
                row['Name'] += '*'
            vals = (
                row['Name'],
                row['SHA'],
                row['Date'],
                row['upstream_sha'],
                row['upstream_date'],
                row['commits_behind'],
            )
            tbl.add_row(vals)
        return tbl

    @classmethod
    def _ptable(cls, fields):
        tbl = pt.PrettyTable(
            field_names=fields,
            border=False,
            hrules=pt.NONE,
            vrules=pt.NONE,
            left_padding_width=2,
            sortby='Name'
        )
        alignments = dict(zip(fields, len(fields) * ['l']))
        tbl.align.update(alignments)
        return tbl



import os
import re
import shutil
import subprocess
from unittest import mock

import pytest

from datakit_project import Create


# override the fixture defined in conftest.py
@pytest.fixture(autouse="module")
def create_plugin_dir():
    pass


def test_missing_config_file(caplog, tmpdir):
    """
    Create should auto-generate config file if it's missing
    """
    cmd = Create(None, None, cmd_name='project create')
    parsed_args = mock.Mock()
    parsed_args.template = ''
    parsed_args.make_default = False
    parsed_args.interactive = False
    cmd.run(parsed_args)
    msg = 'No project templates have been installed'
    assert msg in caplog.text
    assert cmd.configs == {'default_template': ''}
    assert os.path.exists(cmd.plugin_config_path)




import logging
import os

from datakit.utils import (
    home_dir,
    mkdir_p,
    read_json,
    write_json
)


class CommandHelpers:

    "Base class containing common helper methods for Cliff Command instances"

    log = logging.getLogger(__name__)

    def get_template(self, parsed_args):
        if parsed_args.template != '':
            return parsed_args.template
        else:
            return self.default_template

    def update_configs(self, new_configs):
        configs = self.configs
        configs.update(new_configs)
        self.write_configs(configs)
        return configs

    def write_configs(self, configs):
        mkdir_p(self.plugin_config_parent_dir)
        write_json(self.plugin_config_path, configs)

    @property
    def configs(self):
        try:
            configs = read_json(self.plugin_config_path)
        except FileNotFoundError:
            configs = self.default_configs
            self.write_configs(configs)
        return configs

    @property
    def plugin_config_parent_dir(self):
        return os.path.join(
            home_dir(),
            'plugins/datakit-project'
        )

    @property
    def plugin_config_path(self):
        return os.path.join(
            self.plugin_config_parent_dir,
            'config.json'
        )

    @property
    def default_configs(self):
        return {'default_template': ''}

    @property
    def default_template(self):
        try:
            return self.configs['default_template']
        except KeyError:
            msg = "ERROR: The 'default_template' setting is missing from" +\
                "your config file! {}".format(self.plugin_config_path)
            self.log.info(msg)



TEMPLATE_INSTALL_ON_CREATE_MSG = """
To install a new template, you must invoke "datakit project create" with
the local path or URL to a Cookiecutter project repo, e.g.:

    datakit project create --template /local/path/to/my-awesome-cookiecutter-template

"""

NO_TEMPLATES_ERROR_MSG = '\n'.join(("No project templates have been installed!", TEMPLATE_INSTALL_ON_CREATE_MSG))

TEMPLATE_STATUS_MSG = """Use the below command for more detailed status information on templates:

    datakit project templates --status
"""

TEMPLATE_UPDATE_MSG = """To update locally installed Cookiecutter templates:

    datakit project templates update

"""

TEMPLATE_INSTALL_HELP_MSG = """Check out the below URLs for pre-baked templates and installation options:

  - A Pantry Full of Cookiecutters
      https://cookiecutter.readthedocs.io/en/latest/readme.html#a-pantry-full-of-cookiecutters

  - Template install docs
      https://datakit-project.readthedocs.io/en/latest/#creating-your-first-project
"""

TEMPLATE_USAGE_MSG = """To use a locally installed template:

    datakit project create --template my-awesome-cookiecutter
"""

NO_TEMPLATES_ERROR_WITH_HELP_MSG = '\n'.join((NO_TEMPLATES_ERROR_MSG, TEMPLATE_INSTALL_HELP_MSG))

CREATE_HELP_MSG = '\n'.join((
    TEMPLATE_USAGE_MSG,
    TEMPLATE_INSTALL_ON_CREATE_MSG,
    TEMPLATE_INSTALL_HELP_MSG
))



import subprocess

from pathlib import Path

from datakit_project.exceptions import RepositoryFetchFailed


class Repository:

    def __init__(self, path):
        self.path = Path(path)

    def info(self):
        cmd = self._log_pretty_cmd()
        output = self._syscall(cmd)
        return self._prepare_repo_info(output.strip('"'))

    def upstream_info(self):
        cmd = self._log_pretty_cmd()
        cmd.append(
            self.upstream_tracking_branch
        )
        output = self._syscall(cmd)
        info = self._prepare_repo_info(output.strip('"'))
        return {
            'upstream_sha': info['SHA'],
            'upstream_date': info['Date'],
            'upstream_subject': info['Subject'],
            'commits_behind': self.commits_behind_upstream,
        }

    def update(self):
        remote, branch = self.upstream_tracking_branch.split('/')
        cmd = ['git', 'pull', remote, branch]
        self._syscall(cmd)

    @property
    def local_branch(self):
        cmd = ['git', 'rev-parse', '--abbrev-ref', 'HEAD']
        return self._syscall(cmd)

    @property
    def upstream_tracking_branch(self):
        cmd = ['git', 'rev-parse', '--abbrev-ref', 'HEAD@{upstream}']
        return self._syscall(cmd)

    @property
    def commits_behind_upstream(self):
        cmd = ['git', 'rev-list', '--count', "{}..{}".format(
            self.local_branch,
            self.upstream_tracking_branch
        )]
        return int(self._syscall(cmd))

    def fetch(self):
        cmd = ['git', 'fetch']
        self._syscall(cmd)

    def _log_pretty_cmd(self):
        cmd = ['git', 'log', '-n', '1', '--pretty=format:"%h\n%cd\n%s"', '--date=short']
        return cmd

    def _prepare_repo_info(self, info):
        sha1, commit_date, subject = info.split('\n')
        return {
            'Name': self.path.name,
            'SHA': sha1,
            'Date': commit_date,
            'Subject': subject
        }

    def _syscall(self, cmd):
        try:
            output = subprocess.check_output(
                cmd,
                cwd=str(self.path),
                stderr=subprocess.STDOUT
            ).decode().strip()
            return output
        except subprocess.CalledProcessError as git_error:
            output = git_error.output.decode('utf-8')
            if 'fatal: unable to access' in output:
                msg = [output.split("fatal:")[1].strip()]
                msg.append("Are you connected to a network?")
                final = "\n".join(msg)
                raise RepositoryFetchFailed(final)



import os
import re
import shutil
import subprocess
from unittest import mock

import pytest

from datakit_project import Create


@pytest.mark.usefixtures('create_plugin_config_default')
def test_warning_when_no_default_repo_setting(caplog):
    """
    Test CLI warning when 'project create' invoked for the first time
    without specifying a template
    """
    cmd = Create(None, None, cmd_name='project create')
    parsed_args = mock.Mock()
    parsed_args.template = ''
    parsed_args.make_default = False
    parsed_args.interactive = False
    cmd.run(parsed_args)
    msg = 'No project templates have been installed'
    assert msg in caplog.text


@pytest.mark.usefixtures('create_plugin_config_default')
def test_create_initial_project(caplog, monkeypatch, tmpdir):
    """
    Test creation of first project with specification of a
    local Cookiecutter template
    """
    fake_repo_path = os.path.join(os.getcwd(), 'tests/fake-repo')
    # Switch directories
    monkeypatch.chdir(tmpdir)
    # Run the command
    parsed_args = mock.Mock()
    parsed_args.template = fake_repo_path
    parsed_args.make_default = False
    parsed_args.interactive = False
    cmd = Create(None, None, cmd_name='project create')
    cmd.run(parsed_args)
    # Tests:
    #  Project was created
    assert 'fake-project' in [pth.basename for pth in tmpdir.listdir()]
    #  Template set as default since it's first template encountered
    assert cmd.configs['default_template'] == fake_repo_path
    msg_pattern = r"Set default template to.+?fake-repo in plugin config (.+?config.json)"
    log_pattern_matches = True if re.search(msg_pattern, caplog.text) else False
    assert log_pattern_matches


@pytest.mark.usefixtures('create_plugin_config_fake_repo')
def test_usage_of_default_template(monkeypatch, tmpdir):
    """
    Create should use the default template if a new template is not specified.
    """
    # Set up the configs to point default template at our fake repo
    monkeypatch.chdir(tmpdir)
    parsed_args = mock.Mock()
    # NOTE: We're NOT specifying a template here to
    # test usage of default set in fixture
    parsed_args.template = ''
    parsed_args.make_default = False
    parsed_args.interactive = False
    cmd = Create(None, None, cmd_name='project create')
    cmd.run(parsed_args)
    assert 'fake-project' in [pth.basename for pth in tmpdir.listdir()]


@pytest.mark.usefixtures('create_plugin_config_fake_repo')
def test_graceful_handling_of_project_overwrite_attempt(monkeypatch, tmpdir, caplog):
    """
    Create should exit gracefully and provide an appropriate
    error message when user provides a slug with the same name
    as an existing folder in the current working directory.
    """
    # Set up the configs to point default template at our fake repo
    monkeypatch.chdir(tmpdir)
    parsed_args = mock.Mock()
    parsed_args.template = ''
    parsed_args.make_default = False
    parsed_args.interactive = False
    parsed_args.overwrite_if_exists = False
    cmd = Create(None, None, cmd_name='project create')
    cmd.run(parsed_args)
    cmd.run(parsed_args)
    assert 'Error: A project with the slug you provided already exists in this directory. Try again with a different slug.' in caplog.text


@pytest.mark.usefixtures('create_plugin_config_fake_repo')
def test_usage_of_nondefault_template(monkeypatch, tmpdir):
    """
    Create should support use of non-default template
    without overriding default template setting.
    """
    original_repo = os.path.join(os.getcwd(), 'tests/fake-repo')
    new_repo = os.path.join(os.getcwd(), 'tests/fake-repo-two')
    monkeypatch.chdir(tmpdir)
    parsed_args = mock.Mock()
    parsed_args.template = new_repo
    parsed_args.make_default = False
    parsed_args.interactive = False
    cmd = Create(None, None, cmd_name='project create')
    cmd.run(parsed_args)
    dir_contents = [pth.basename for pth in tmpdir.listdir()]
    assert 'fake-project' not in dir_contents
    assert 'fake-project-two' in dir_contents
    assert cmd.configs['default_template'] == original_repo


@pytest.mark.usefixtures('create_plugin_config_fake_repo')
def test_new_default_template(caplog, monkeypatch, tmpdir):
    """
    Create should support ability to set a new default template.
    """
    original_repo = os.path.join(os.getcwd(), 'tests/fake-repo')
    new_repo = os.path.join(os.getcwd(), 'tests/fake-repo-two')
    monkeypatch.chdir(tmpdir)
    # On first pass, we specify new template and set it as new default
    parsed_args = mock.Mock()
    parsed_args.template = new_repo
    parsed_args.make_default = True
    parsed_args.interactive = False
    cmd = Create(None, None, cmd_name='project create')
    cmd.run(parsed_args)
    dir_contents = [pth.basename for pth in tmpdir.listdir()]
    assert 'fake-project' not in dir_contents
    assert 'fake-project-two' in dir_contents
    assert cmd.configs['default_template'].endswith('fake-repo-two')
    msg_pattern = r"Set default template to.+?fake-repo-two in plugin config (.+?config.json)"
    log_pattern_matches = True if re.search(msg_pattern, caplog.text) else False
    assert log_pattern_matches
    # Remove fake project and recreate without specifying the template
    # to ensure it's now applied as the default
    project_directory = os.path.join(tmpdir.strpath, 'fake-project-two')
    shutil.rmtree(project_directory)
    args_new = mock.Mock()
    args_new.template = ''
    args_new.interactive = False
    cmd = Create(None, None, cmd_name='project create')
    cmd.run(args_new)
    dir_contents_updated = [pth.basename for pth in tmpdir.listdir()]
    assert 'fake-project' not in dir_contents_updated
    assert 'fake-project-two' in dir_contents_updated


def test_github_install(caplog, monkeypatch, cookiecutter_home, tmpdir):
    orig_repo = os.path.join(os.getcwd(), 'tests/fake-repo')
    new_repo = os.path.join(cookiecutter_home, 'fake-repo')

    # Monkeypatch subprocess to prevent actual git clone
    # and simply copy over the repo instead
    def mockreturn(repo_bits, cwd, stderr):
        shutil.copytree(orig_repo, new_repo)
        return 'gh:associatedpress/fake-repo'

    monkeypatch.setattr(subprocess, 'check_output', mockreturn)
    monkeypatch.chdir(tmpdir)
    parsed_args = mock.Mock()
    # NOTE: We're NOT specifying a template here, but we set the default above
    parsed_args.template = 'gh:associatedpress/fake-repo'
    parsed_args.make_default = False
    parsed_args.interactive = False
    cmd = Create(None, None, cmd_name='project create')
    cmd.run(parsed_args)
    assert 'fake-repo' in os.listdir(cookiecutter_home)
    assert cmd.configs['default_template'] == new_repo



import argparse
from cliff.command import Command
from cookiecutter.main import cookiecutter
from cookiecutter.prompt import read_user_choice
from cookiecutter.exceptions import OutputDirExistsException
import cookiecutter.config as cc_config

from .command_helpers import CommandHelpers
from .help_text import CREATE_HELP_MSG, NO_TEMPLATES_ERROR_WITH_HELP_MSG
from datakit_project.cookiecutters import Cookiecutters
from datakit_project.utils import resolve_repo_dir


class Create(CommandHelpers, Command):
    """Create a new project."""

    def get_epilog(self):
        return CREATE_HELP_MSG

    def get_parser(self, prog_name):
        parser = super(Create, self).get_parser(prog_name)
        parser.formatter_class = argparse.RawTextHelpFormatter
        parser.add_argument(
            '-t',
            '--template',
            default='',
            help="Local path/Github URL for a Cookiecutter template"
        )
        parser.add_argument(
            '-d',
            '--make-default',
            action='store_true',
            default=False,
            help="Make the template the default for new projects"
        )
        parser.add_argument(
            '-i',
            '--interactive',
            action='store_true',
            default=False,
            help="Enter interactive mode to choose a project template"
        )
        parser.add_argument(
            '-n',
            '--no-input',
            action='store_true',
            default=False,
            help="Whether to allow user input for Cookiecutter execution"
        )
        parser.add_argument(
            '-o',
            '--overwrite-if-exists',
            action='store_true',
            default=False,
            help="Whether to overwrite a directory with the same name as project slug or exit gracefully"
        )
        parser.add_argument(
            '-c',
            '--checkout',
            action='store',
            help="Branch, tag or commit to checkout from git clone"
        )
        return parser

    def take_action(self, parsed_args):
        if parsed_args.interactive and parsed_args.template == '':
            cc_home = cc_config.DEFAULT_CONFIG['cookiecutters_dir']
            cc = Cookiecutters(cc_home)
            templates = cc.list_templates()
            template = read_user_choice('project template', templates)
        else:
            # Create project skeleton
            template = self.get_template(parsed_args)
        if template:
            self.log.info("Creating project from template: {}".format(template))
            try:
                cookiecutter(
                    template,
                    checkout=parsed_args.checkout,
                    no_input=parsed_args.no_input,
                    overwrite_if_exists=parsed_args.overwrite_if_exists
                )
                repo_dir = resolve_repo_dir(template)
                # Update default template if it's empty or specifically requested
                if self.default_template == '' or parsed_args.make_default:
                    self.update_configs({'default_template': repo_dir})
                    tmplt_msg = "Set default template to {} in plugin config ({})".format(
                        repo_dir,
                        self.plugin_config_path
                    )
                    self.log.info(tmplt_msg)
            except OutputDirExistsException:
                msg = (
                    "Error: A project with the slug you provided already "
                    "exists in this directory. Try again with a different slug."
                )
                self.log.info(msg)
        else:
            self.log.info(NO_TEMPLATES_ERROR_WITH_HELP_MSG)



"""

datakit-project
---------------

A project-skeleton generator for use with the `datakit <https://pypi.python.org/pypi/datakit-core/>`_ framework.

* `Code <https://github.com/associatedpress/datakit-project>`_
* `Docs <http://datakit-project.readthedocs.io/en/latest/>`_

"""
from setuptools import setup, find_packages

PROJECT = 'datakit-project'
VERSION = '0.5.0'

setup(
    name=PROJECT,
    version=VERSION,
    description="A datakit plugin to generate new projects.",
    long_description=__doc__,
    long_description_content_type='text/x-rst',
    author="Serdar Tumgoren",
    author_email='zstumgoren@gmail.com',
    license="ISCL",
    url='https://datakit.ap.org',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='datakit',
    platforms=['Any'],
    install_requires=[
        'click',
        'cliff',
        'cookiecutter>=1.6.0',
        'datakit-core',
        'prettytable'
    ],
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        'datakit.plugins': [
            'project create = datakit_project:Create',
            'project templates = datakit_project:Templates',
            'project templates update = datakit_project:TemplatesUpdate',
        ]
    },
    test_suite='tests',
    tests_require=[
        'pytest',
        'pytest-catchlog',
    ],
    project_urls={
        'Maintainer': 'https://github.com/associatedpress',
        'Source': 'https://github.com/associatedpress/datakit-project',
        'Tracker': 'https://github.com/associatedpress/datakit-project/issues'
    },
    zip_safe=False,
)



import os
import re
import shutil
import subprocess
from unittest import mock

import cookiecutter.config as cc_config
import pytest

from datakit_project import cookiecutters, Templates
from datakit_project import exceptions as dkit_exceptions


def test_noupdate_status(caplog, cookiecutter_home, deploy_template, monkeypatch, tmpdir):
    deploy_template(cookiecutter_home, 'tests/fake-repo')
    deploy_template(cookiecutter_home, 'tests/fake-repo-two')
    # Switch directories
    monkeypatch.chdir(tmpdir)
    # Mock the return value for Cookiecutters.info
    with mock.patch('datakit_project.commands.templates.Cookiecutters') as MockClass:
        instance = MockClass.return_value
        instance.info.return_value = [
            {
              'Name': 'fake-repo',
              'SHA': 'a11706f',
              'Date': '2017-02-02',
              'Subject': 'Testing',
              'upstream_sha': 'a11706f',
              'upstream_date': '2017-02-02',
              'upstream_subject': 'Testing',
              'commits_behind': 'Up-to-date'
            },
            {
              'Name': 'fake-repo-two',
              'SHA': '55f62a0',
              'Date': '2019-03-06',
              'Subject': 'Testing',
              'upstream_sha': '55f62a0',
              'upstream_date': '2019-03-06',
              'upstream_subject': 'Testing',
              'commits_behind': 'Up-to-date'
            },
        ]
        # Run the command
        parsed_args = mock.Mock()
        # Set status flag to true
        parsed_args.status = True
        cmd = Templates(None, None, cmd_name='project templates')
        cmd.run(parsed_args)
        header = "Name            SHA       Date         Upstream SHA   Upstream Date   Commits behind"
        repo1 = "fake-repo       a11706f   2017-02-02   a11706f        2017-02-02      Up-to-date"
        repo2 = "fake-repo-two   55f62a0   2019-03-06   55f62a0        2019-03-06      Up-to-date"
        expected = [header, repo1, repo2]
        # Some gross cleanup of caplog.text is necessary because it displays log level info
        actual = [re.sub(r"templates.py\s+\d+\sINFO", "", line).strip()  for line in caplog.text.split('\n')]
        # Test that expected lines appear in displayed text
        for line in expected:
            assert line in actual
        # Test alpha sortting by repo name
        header_idx = actual.index(header)
        repo1_idx = actual.index(repo1)
        repo2_idx = actual.index(repo2)
        assert header_idx < repo1_idx
        assert repo1_idx < repo2_idx


def test_repo_behind_upstream(caplog, cookiecutter_home, deploy_template, monkeypatch, tmpdir):
    deploy_template(cookiecutter_home, 'tests/fake-repo')
    deploy_template(cookiecutter_home, 'tests/fake-repo-two')
    # Switch directories
    monkeypatch.chdir(tmpdir)
    # Mock the return value for Cookiecutters.info
    with mock.patch('datakit_project.commands.templates.Cookiecutters') as MockClass:
        instance = MockClass.return_value
        instance.info.return_value = [
            {
              'Name': 'fake-repo',
              'SHA': 'a11706f',
              'Date': '2017-02-02',
              'Subject': 'Testing',
              'upstream_sha': 'c18705e',
              'upstream_date': '2017-06-01',
              'upstream_subject': 'Some newer commit',
              'commits_behind': 3
            },
            {
              'Name': 'fake-repo-two',
              'SHA': '55f62a0',
              'Date': '2019-03-06',
              'Subject': 'Testing',
              'upstream_sha': '55f62a0',
              'upstream_date': '2019-03-06',
              'upstream_subject': 'Testing',
              'commits_behind': 'Up-to-date'
            },
        ]
        # Run the command
        parsed_args = mock.Mock()
        # Set status flag to true
        parsed_args.status = True
        cmd = Templates(None, None, cmd_name='project templates')
        cmd.run(parsed_args)
        header = "Name            SHA       Date         Upstream SHA   Upstream Date   Commits behind"
        repo1 = "fake-repo       a11706f   2017-02-02   c18705e        2017-06-01      3"
        repo2 = "fake-repo-two   55f62a0   2019-03-06   55f62a0        2019-03-06      Up-to-date"
        expected = [header, repo1, repo2]
        # Some gross cleanup of caplog.text is necessary because it displays log level info
        actual = [re.sub(r"templates.py\s+\d+\sINFO", "", line).strip()  for line in caplog.text.split('\n')]
        # Test that expected lines appear in displayed text
        for line in expected:
            assert line in actual
        # Test alpha sortting by repo name
        header_idx = actual.index(header)
        repo1_idx = actual.index(repo1)
        repo2_idx = actual.index(repo2)
        assert header_idx < repo1_idx
        assert repo1_idx < repo2_idx
